from .simple_embeddings import (
    Angular2DEmbeddings,
    AnimalEmbeddings,
    EarthEmbeddings,
    ParserEmbeddings,
)

__all__ = [
    "Angular2DEmbeddings",
    "AnimalEmbeddings",
    "EarthEmbeddings",
    "ParserEmbeddings",
]
